@@ -2,7 +2,7 @@
 
     Main file for  devfsd  (devfs daemon for Linux).
 
-    Copyright (C) 1998-1999  Richard Gooch
+    Copyright (C) 1998-2000  Richard Gooch
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -114,9 +114,15 @@
     Updated by      Richard Gooch   17-DEC-1999: Do not grab event queue when
   just setting devfs debug mask.
 
-    Last updated by Richard Gooch   20-DEC-1999: Swap pathname and action
+    Updated by      Richard Gooch   20-DEC-1999: Swap pathname and action
   fields in config file.
 
+    Updated by      Richard Gooch   25-DEC-1999: Discard lookups on "/dev/log"
+  and "/dev/initctl". Added "MODLOAD" action.
+
+    Last updated by Richard Gooch   3-FEB-2000: Bitch and moan if modprobe(8)
+  is used with "EXECUTE" action.
+
 
 */
 #include <unistd.h>
@@ -140,6 +146,7 @@
 #include <signal.h>
 #include <regex.h>
 #include <errno.h>
+#include <dlfcn.h>
 #include <karma.h>
 
 #ifndef IDE6_MAJOR        /*  In case we're building with an ancient kernel  */
@@ -150,7 +157,7 @@
 #endif
 
 
-#define VERSION "1.3.0"
+#define VERSION "1.3.1"
 #define CONFIG_FILE "/etc/devfsd.conf"
 #define MAX_ARGS (6 + 1)
 
@@ -164,12 +171,13 @@
 
 
 #define AC_PERMISSIONS              0
-#define AC_EXECUTE                  1
-#define AC_IGNORE                   2
-#define AC_MKOLDCOMPAT              3
-#define AC_MKNEWCOMPAT              4
-#define AC_RMOLDCOMPAT              5
-#define AC_RMNEWCOMPAT              6
+#define AC_MODLOAD                  1
+#define AC_EXECUTE                  2
+#define AC_IGNORE                   3
+#define AC_MKOLDCOMPAT              4
+#define AC_MKNEWCOMPAT              5
+#define AC_RMOLDCOMPAT              6
+#define AC_RMNEWCOMPAT              7
 
 #define MKACTION(what,when) (struct action_type) {what, when}
 
@@ -233,6 +241,8 @@ static void do_servicing (int fd, unsigned long event_mask);
 static void service_name (const struct devfsd_notify_struct *info);
 static void action_permissions (const struct devfsd_notify_struct *info,
 				const struct config_entry *entry);
+static void action_modload (const struct devfsd_notify_struct *info,
+			    const struct config_entry *entry);
 static void action_execute (const struct devfsd_notify_struct *info,
 			    const struct config_entry *entry);
 static void action_compat (const struct devfsd_notify_struct *info,
@@ -553,10 +563,15 @@ static void read_config (CONST char *filename, flag optional,
 	    /*  Get mode  */
 	    new->u.permissions.mode = get_mode (p[1]);
 	}
+	else if (strcasecmp (what, "MODLOAD") == 0)
+	    new->action.what = AC_MODLOAD;
 	else if (strcasecmp (what, "EXECUTE") == 0)
 	{
 	    new->action.what = AC_EXECUTE;
 	    num_args -= 3;
+	    if (strstr (p[0], "modprobe") != NULL)  /*  Bitch and moan  */
+		SYSLOG (LOG_NOTICE,
+			"PLEASE USE MODLOAD ACTION INSTEAD OF EXECUTE modprobe\n");
 	    for (count = 0; count < num_args; ++count)
 	    {
 		err = strlen (p[count]) + 1;
@@ -675,11 +690,20 @@ static void service_name (const struct devfsd_notify_struct *info)
 	 */
 	if (info->type != entry->action.when) continue;
 	if (regexec (&entry->preg, info->devname, 1, &mbuf, 0) != 0) continue;
+	/*  Discard lookups on "/dev/log" and "/dev/initctl"  */
+	if (info->type == DEVFSD_NOTIFY_LOOKUP)
+	{
+	    if ( (strcmp (info->devname, "log") == 0) ||
+		 (strcmp (info->devname, "initctl") == 0) ) continue;
+	}
 	switch (entry->action.what)
 	{
 	  case AC_PERMISSIONS:
 	    action_permissions (info, entry);
 	    break;
+	  case AC_MODLOAD:
+	    action_modload (info, entry);
+	    break;
 	  case AC_EXECUTE:
 	    action_execute (info, entry);
 	    break;
@@ -755,6 +779,58 @@ static void action_permissions (const struct devfsd_notify_struct *info,
     }
 }   /*  End Function action_permissions  */
 
+static void action_modload (const struct devfsd_notify_struct *info,
+			    const struct config_entry *entry)
+/*  [SUMMARY] Load a module.
+    <info> The devfs change.
+    <entry> The config file entry.
+    [RETURNS] Nothing.
+*/
+{
+    char *argv[4];
+    char device[STRING_LENGTH];
+    static int (*modprobe) (int argc, char **argv) = NULL;
+    static flag first_time = TRUE;
+
+    if (first_time)
+    {
+	void *lib;
+
+	first_time = FALSE;
+	if ( ( lib = dlopen ("/lib/modutils.so", RTLD_NOW) ) != NULL )
+	    modprobe = dlsym (lib, "modprobe");
+    }
+    argv[0] = "/sbin/modprobe";
+    argv[1] = "-k";
+    argv[2] = device;
+    argv[3] = NULL;
+    strcpy (device, "/dev/");
+    strcpy (device + 5, info->devname);
+    if (modprobe != NULL)
+    {
+	(*modprobe) (3, argv);
+	return;
+    }
+    switch ( fork () )
+    {
+      case 0:
+	/*  Child  */
+	break;
+      case -1:
+	/*  Error  */
+	SYSLOG (LOG_ERR, "error forking\t%s\n", ERRSTRING);
+	exit (2);
+	/*break;*/
+      default:
+	/*  Parent  */
+	wait (NULL);
+	return;
+	/*break;*/
+    }
+    execvp (argv[0], argv);
+    SYSLOG (LOG_ERR, "error execing: \"%s\"\t%s\n", argv[0], ERRSTRING);
+}   /*  End Function action_modload  */
+
 static void action_execute (const struct devfsd_notify_struct *info,
 			    const struct config_entry *entry)
 /*  [SUMMARY] Execute a programme.
